package aQute.bnd.plugin.editors;

import org.eclipse.jface.text.*;
import org.eclipse.jface.text.presentation.*;
import org.eclipse.jface.text.rules.*;
import org.eclipse.jface.text.source.*;

public class BndSourceViewerConfiguration extends SourceViewerConfiguration {
	static final String	SINGLELINE_COMMENT_TYPE	= "___slc";

	
	class BndPartitionScanner extends RuleBasedPartitionScanner {
		public BndPartitionScanner() {
			IToken singleLineComment = new Token(SINGLELINE_COMMENT_TYPE);

			IPredicateRule[] rules = new IPredicateRule[] { new EndOfLineRule(
					"#", singleLineComment), };
			setPredicateRules(rules);
		}
	}

	private BndScanner					_scanner;
	private MultiLineCommentScanner		_multiLineCommentScanner;
	private ISharedTextColors			_colorManager;

	public BndSourceViewerConfiguration(ISharedTextColors colors) {
		this._colorManager = colors;
	}

	public IPresentationReconciler getPresentationReconciler(
			ISourceViewer sourceViewer) {
		PresentationReconciler reconciler = new PresentationReconciler();
		configureReconciler(reconciler, IDocument.DEFAULT_CONTENT_TYPE,
				getBndScanner());
		configureReconciler(reconciler, SINGLELINE_COMMENT_TYPE,
				getMultiLineCommentScanner());
		return reconciler;
	}

	private void configureReconciler(PresentationReconciler reconciler,
			String partitionType, ITokenScanner scanner) {
		DefaultDamagerRepairer dr;
		dr = new DefaultDamagerRepairer(scanner);
		reconciler.setDamager(dr, partitionType);
		reconciler.setRepairer(dr, partitionType);
	}

	protected BndScanner getBndScanner() {
		if (_scanner == null) {
			_scanner = new BndScanner(_colorManager);
		}
		return _scanner;
	}

	class MultiLineCommentScanner extends RuleBasedScanner {

		public MultiLineCommentScanner(ISharedTextColors manager) {
			IToken comment = new Token(new TextAttribute(manager
					.getColor(BndScanner.C_COMMENT)));
			setDefaultReturnToken(comment);
		}

	}

	protected MultiLineCommentScanner getMultiLineCommentScanner() {
		if (_multiLineCommentScanner == null) {
			_multiLineCommentScanner = new MultiLineCommentScanner(
					_colorManager);
		}
		return _multiLineCommentScanner;
	}


	public String[] getDefaultPrefixes(ISourceViewer sourceViewer,
			String contentType) {
		if (IDocument.DEFAULT_CONTENT_TYPE == contentType ||
				SINGLELINE_COMMENT_TYPE == contentType) {
			return new String[] { "#", "//" };
		}
		return null;
	}

}